@@ -13,13 +13,14 @@ import pandas as pd
 import tensorflow.compat.v1 as tf
 from tensorflow.keras.initializers import truncated_normal
 from .base import Base, TfMixin
+from ..data.data_generator import DataGenFeat
 from ..evaluation.evaluate import EvalMixin
 from ..utils.tf_ops import (
     reg_config,
     dropout_config,
-    lr_decay_config
+    lr_decay_config,
+    multi_sparse_combine_embedding
 )
-from ..data.data_generator import DataGenFeat
 from ..utils.sampling import NegativeSampling
 from ..feature import (
     get_predict_indices_and_values,
@@ -54,6 +55,7 @@ class AutoInt(Base, TfMixin, EvalMixin):
             dropout_rate=None,
             hidden_units="128,64,32",
             batch_sampling=False,
+            multi_sparse_combiner="sqrtn",
             seed=42,
             lower_upper_bound=None,
             tf_sess_config=None
@@ -90,6 +92,10 @@ class AutoInt(Base, TfMixin, EvalMixin):
         if self.sparse:
             self.sparse_feature_size = self._sparse_feat_size(data_info)
             self.sparse_field_size = self._sparse_field_size(data_info)
+            self.multi_sparse_combiner = self._check_multi_sparse(
+                data_info, multi_sparse_combiner)
+            self.true_sparse_field_size = self._true_sparse_field_size(
+                data_info, self.sparse_field_size, self.multi_sparse_combiner)
         if self.dense:
             self.dense_field_size = self._dense_field_size(data_info)
         self.all_args = locals()
@@ -149,7 +155,14 @@ class AutoInt(Base, TfMixin, EvalMixin):
             initializer=truncated_normal(0.0, 0.01),
             regularizer=self.reg)
 
-        sparse_embed = tf.nn.embedding_lookup(sparse_feat, self.sparse_indices)
+        if (self.data_info.multi_sparse_combine_info
+                and self.multi_sparse_combiner in ("sum", "mean", "sqrtn")):
+            sparse_embed = multi_sparse_combine_embedding(
+                self.data_info, sparse_feat, self.sparse_indices,
+                self.multi_sparse_combiner, self.embed_size)
+        else:
+            sparse_embed = tf.nn.embedding_lookup(sparse_feat, self.sparse_indices)
+
         self.concat_embed.append(sparse_embed)
 
     def _build_dense(self):
