@@ -7,14 +7,15 @@ author: massquantity
 
 """
 from itertools import islice
+import os
 import numpy as np
-import tensorflow as tf2
+import tensorflow.compat.v1 as tf
 from tensorflow.keras.initializers import (
     truncated_normal as tf_truncated_normal,
     glorot_normal as tf_glorot_normal
 )
 from .base import Base, TfMixin
-from ..evaluate.evaluate import EvalMixin
+from ..evaluation.evaluate import EvalMixin
 from ..utils.tf_ops import (
     reg_config,
     dropout_config,
@@ -23,13 +24,17 @@ from ..utils.tf_ops import (
 from ..data.data_generator import DataGenSequence
 from ..data.sequence import user_last_interacted
 from ..utils.misc import time_block, colorize
-from ..utils.misc import count_params
+from ..utils.misc import count_params, assign_oov_vector
 from ..utils.tf_ops import conv_nn, max_pool
-tf = tf2.compat.v1
 tf.disable_v2_behavior()
 
 
 class Caser(Base, TfMixin, EvalMixin):
+    user_variables = ["user_feat"]
+    item_variables = ["item_weights", "item_biases", "input_embed"]
+    user_variables_np = ["user_vector"]
+    item_variables_np = ["item_vector"]
+
     def __init__(
             self,
             task,
@@ -53,7 +58,7 @@ class Caser(Base, TfMixin, EvalMixin):
     ):
         Base.__init__(self, task, data_info, lower_upper_bound)
         TfMixin.__init__(self, tf_sess_config)
-        EvalMixin.__init__(self, task)
+        EvalMixin.__init__(self, task, data_info)
 
         self.task = task
         self.data_info = data_info
@@ -82,8 +87,11 @@ class Caser(Base, TfMixin, EvalMixin):
         self.item_vector = None
         self.sparse = False
         self.dense = False
+        self.vector_infer = True
+        self.all_args = locals()
 
     def _build_model(self):
+        self.graph_built = True
         tf.set_random_seed(self.seed)
         self._build_placeholders()
         self._build_variables()
@@ -193,7 +201,7 @@ class Caser(Base, TfMixin, EvalMixin):
         )
         self.user_embed = tf.concat([user_repr, convs_out], axis=1)
 
-    def _build_train_ops(self, global_steps=None):
+    def _build_train_ops(self, **kwargs):
         if self.task == "rating":
             self.loss = tf.losses.mean_squared_error(labels=self.labels,
                                                      predictions=self.output)
@@ -209,6 +217,13 @@ class Caser(Base, TfMixin, EvalMixin):
         else:
             total_loss = self.loss
 
+        if self.lr_decay:
+            n_batches = int(self.data_info.data_size / self.batch_size)
+            self.lr, global_steps = lr_decay_config(self.lr, n_batches,
+                                                    **kwargs)
+        else:
+            global_steps = None
+
         optimizer = tf.train.AdamOptimizer(self.lr)
         optimizer_op = optimizer.minimize(total_loss, global_step=global_steps)
         self.training_op = optimizer_op
@@ -217,16 +232,9 @@ class Caser(Base, TfMixin, EvalMixin):
     def fit(self, train_data, verbose=1, shuffle=True,
             eval_data=None, metrics=None, **kwargs):
         self.show_start_time()
-        if self.lr_decay:
-            n_batches = int(len(train_data) / self.batch_size)
-            self.lr, global_steps = lr_decay_config(
-                self.lr, n_batches, **kwargs
-            )
-        else:
-            global_steps = None
-
-        self._build_model()
-        self._build_train_ops(global_steps)
+        if not self.graph_built:
+            self._build_model()
+            self._build_train_ops(**kwargs)
 
         data_generator = DataGenSequence(
             data=train_data,
@@ -264,23 +272,16 @@ class Caser(Base, TfMixin, EvalMixin):
                 print(f"\t {colorize(train_loss_str, 'green')}")
                 # for evaluation
                 self._set_latent_factors()
-                self.print_metrics(eval_data=eval_data, metrics=metrics)
-                print("=" * 30)
+                self.print_metrics(eval_data=eval_data, metrics=metrics,
+                                   **kwargs)
+                print("="*30)
 
         # for prediction and recommendation
         self._set_latent_factors()
+        assign_oov_vector(self)
 
-    def predict(self, user, item):
-        user = (
-            np.asarray([user])
-            if isinstance(user, int)
-            else np.asarray(user)
-        )
-        item = (
-            np.asarray([item])
-            if isinstance(item, int)
-            else np.asarray(item)
-        )
+    def predict(self, user, item, cold_start="average", inner_id=False):
+        user, item = self.convert_id(user, item, inner_id)
         unknown_num, unknown_index, user, item = self._check_unknown(user, item)
 
         preds = np.sum(
@@ -294,29 +295,34 @@ class Caser(Base, TfMixin, EvalMixin):
         elif self.task == "ranking":
             preds = 1 / (1 + np.exp(-preds))
 
-        if unknown_num > 0:
+        if unknown_num > 0 and cold_start == "popular":
             preds[unknown_index] = self.default_prediction
+        return preds
 
-        return preds[0] if len(user) == 1 else preds
+    def recommend_user(self, user, n_rec, cold_start="average", inner_id=False):
+        user_id = self._check_unknown_user(user, inner_id)
+        if user_id is None:
+            if cold_start == "average":
+                user_id = self.n_users
+            elif cold_start == "popular":
+                return self.data_info.popular_items[:n_rec]
+            else:
+                raise ValueError(user)
 
-    def recommend_user(self, user, n_rec, **kwargs):
-        user = self._check_unknown_user(user)
-        if not user:
-            return
+        consumed = set(self.user_consumed[user_id])
+        count = n_rec + len(consumed)
+        recos = self.user_vector[user_id] @ self.item_vector.T
 
-        recos = self.user_vector[user] @ self.item_vector.T
         if self.task == "ranking":
             recos = 1 / (1 + np.exp(-recos))
-
-        consumed = self.user_consumed[user]
-        count = n_rec + len(consumed)
         ids = np.argpartition(recos, -count)[-count:]
         rank = sorted(zip(ids, recos[ids]), key=lambda x: -x[1])
-        return list(
-            islice(
-                (rec for rec in rank if rec[0] not in consumed), n_rec
-            )
+        recs_and_scores = islice(
+            (rec if inner_id else (self.data_info.id2item[rec[0]], rec[1])
+             for rec in rank if rec[0] not in consumed),
+            n_rec
         )
+        return list(recs_and_scores)
 
     def _set_last_interacted(self):
         if (self.user_last_interacted is None
@@ -348,3 +354,26 @@ class Caser(Base, TfMixin, EvalMixin):
         item_bias = item_biases[:, None]
         self.user_vector = np.hstack([user_embed, user_bias])
         self.item_vector = np.hstack([item_weights, item_bias])
+
+    def save(self, path, model_name, manual=True, inference_only=False):
+        if not os.path.isdir(path):
+            print(f"file folder {path} doesn't exists, creating a new one...")
+            os.makedirs(path)
+        self.save_params(path)
+        if inference_only:
+            variable_path = os.path.join(path, model_name)
+            np.savez_compressed(variable_path,
+                                user_vector=self.user_vector,
+                                item_vector=self.item_vector)
+        else:
+            self.save_variables(path, model_name, inference_only=False)
+
+    @classmethod
+    def load(cls, path, model_name, data_info, manual=True):
+        variable_path = os.path.join(path, f"{model_name}.npz")
+        variables = np.load(variable_path)
+        hparams = cls.load_params(path, data_info)
+        model = cls(**hparams)
+        model.user_vector = variables["user_vector"]
+        model.item_vector = variables["item_vector"]
+        return model
