@@ -1,5 +1,5 @@
 from .dataset import DatasetPure, DatasetFeat
-from .data_info import DataInfo
+from .data_info import DataInfo, MultiSparseInfo
 from .processing import process_data, split_multi_value
 from .split import (
     split_by_num,
